from nose.tools import assert_equals, assert_true
from tree2tax.threshold_finder import ThresholdFinder
from skbio.tree import TreeNode
from StringIO import StringIO
from string import split as _

class TestThresholdFinder:
    def assertSameCladeDistanceSet(self, expected, observed):
        assert_equals(len(expected), len(observed))
        for i, exp in enumerate(expected):
            assert_equals(exp[0], observed[i].parent_node.name)
            assert_equals(sorted([exp[1], exp[2]]), \
                          sorted([observed[i].daughter_node1.name, observed[i].daughter_node2.name]))
            assert_equals(exp[3], observed[i].distance)
    
    def testSimple(self):
        tree = TreeNode.read(StringIO("(((A:1, B:2)'g__genus1':3, (C:4, D:5)'g__genus2':6)'f__family':10)root;"))
        examples = ThresholdFinder().find_examples(tree, 'f', 'g')
        self.assertSameCladeDistanceSet([['f__family','g__genus1','g__genus2',16.0]],
                                         examples)
    
    def testTreeSubtree(self):
        '''one genus is a subtree of another'''
        tree = TreeNode.read(StringIO("((((A:1, B:2)'g__genus1':3, D:50)'g__genus2':6)'f__family':10)root;"))
        examples = ThresholdFinder().find_examples(tree, 'f', 'g')
        self.assertSameCladeDistanceSet([['f__family','g__genus1','g__genus2',55.0]],
                                         examples)
    def testTreeSubtree2(self):
        '''one genus is a subtree of another, and the longest branch is in both subtrees'''
        tree = TreeNode.read(StringIO("((((A:1, B:52)'g__genus1':3, D:50)'g__genus2':6)'f__family':10)root;"))
        examples = ThresholdFinder().find_examples(tree, 'f', 'g')
        self.assertSameCladeDistanceSet([['f__family','g__genus1','g__genus2',105.0]],
                                         examples)
        
    def testMultiplyNamedNode(self):
        tree = TreeNode.read(StringIO("(((A:1, B:2)'g__genus1':3, (C:4, D:5)'g__genus2; s__spec':6)'f__family':10)root;"))
        examples = ThresholdFinder().find_examples(tree, 'f', 'g')
        self.assertSameCladeDistanceSet([['f__family','g__genus1','g__genus2; s__spec',16.0]],
                                         examples)
        
    def testNoPairs(self):
        tree = TreeNode.read(StringIO("(((A:1, B:2):3, (C:4, D:5):6)'f__family; g__genoos':10)root;"))
        examples = ThresholdFinder().find_examples(tree, 'f', 'g')
        self.assertSameCladeDistanceSet([],
                                         examples)



from nose.tools import assert_equals, assert_true
from tree2tax.tree2tax import Tree2Tax, NamedCluster, TaxonomyFunctions
from skbio.tree import TreeNode
from StringIO import StringIO
from string import split as _

class TestTaxonomyFunctions:
    def test_taxonomy_from_node_name(self):
        assert_equals('C', TaxonomyFunctions().taxonomy_from_node_name('C'))
        assert_equals('C', TaxonomyFunctions().taxonomy_from_node_name('0.997:C'))
        assert_equals(None, TaxonomyFunctions().taxonomy_from_node_name('0.1'))
        
    def test_missing_taxonomy(self):
        tree = TreeNode.read(StringIO('((((A:11, B:12)C:10, D:9)E:20, F:20)G:30)root;'))
        assert_equals(['C'], TaxonomyFunctions().missing_taxonomy(tree, tree.find('A'), tree.find('E')))
        assert_equals([], TaxonomyFunctions().missing_taxonomy(tree, tree.find('A'), tree.find('A')))
        assert_equals(['E','C'], TaxonomyFunctions().missing_taxonomy(tree, tree.find('A'), tree.find('G')))

class TestTree2TaxNamedClusters:
    def assertSameClusters(self, expected, observed):
        name_clusters = [sorted([n.name for n in named_cluster.tips]) for named_cluster in observed]
        expected_name_clusters = [sorted(node_names) for node_names in expected]
        #print "named clusters: %s" % name_clusters
        assert_equals(expected_name_clusters, name_clusters)
         
    def testSimple(self):
        tree = TreeNode.read(StringIO('((A:0.11, B:0.12)C:0.1, D:0.2)root;'))
        clusters = Tree2Tax().named_clusters(tree, 0.25)
        self.assertSameClusters([['A','B'],['D']], clusters)
         
    def testNoClusering(self):
        tree = TreeNode.read(StringIO('((A:0.11, B:0.12)C:0.1, D:0.2)root;'))
        clusters = Tree2Tax().named_clusters(tree, 0.05)
        self.assertSameClusters([['A'],['B'],['D']], clusters)
        assert_equals(_('C.1 C.2 Root'), [c.name() for c in clusters])
          
    def testClusterEverything(self):
        tree = TreeNode.read(StringIO('((A:0.11, B:0.12)C:0.1, D:0.2)root;'))
        clusters = Tree2Tax().named_clusters(tree, 0.5)
        self.assertSameClusters([['A','B','D']], clusters)
        assert_equals('Root',clusters[0].name())
          
    def testClusterOnInternalNode(self):
        tree = TreeNode.read(StringIO('((((A:11, B:12)C:10, D:9)E:20, F:20)G:30)root;'))
        clusters = Tree2Tax().named_clusters(tree, 40)
        self.assertSameClusters([_('A B D'), ['F']], clusters)
          
    def testClusterOnTwoInternalNodes(self):
        tree = TreeNode.read(StringIO('((((A:11, B:12)C:10, (H:8, D:9)I:3)E:20, F:20)G:30)root;'))
        clusters = Tree2Tax().named_clusters(tree, 40)
        self.assertSameClusters([_('A B D H'), ['F']], clusters)
          
    def testClusterIntoThree(self):
        tree = TreeNode.read(StringIO('((((A:11, B:12)C:10, (H:8, D:9)I:3)E:20, F:20)G:30)root;'))
        clusters = Tree2Tax().named_clusters(tree, 25)
        self.assertSameClusters([_('A B'), _('D H'), ['F']], clusters)
         
    def testClusterNamingOnTwoInternalNodes(self):
        tree = TreeNode.read(StringIO('((((A:11, B:12):10, (H:8, D:9):3):20, F:20)G:30)root;'))
        clusters = Tree2Tax().named_clusters(tree, 40)
        self.assertSameClusters([_('A B D H'), ['F']], clusters)
        assert_equals(_('G.1 G.2'), [c.name() for c in clusters])
         
    def testClusterNamingOnTwoInternalNodesReverseOrder(self):
        tree = TreeNode.read(StringIO('((F:20, ((A:11, B:12):10, (H:8, D:9):3):20)G:30)root;'))
        clusters = Tree2Tax().named_clusters(tree, 40)
        self.assertSameClusters([['F'], _('A B D H')], clusters)
        assert_equals(_('G.2 G.1'), [c.name() for c in clusters])
         
    def testNamingWithBootstraps(self):
        tree = TreeNode.read(StringIO('((A:0.11, B:0.12)0.091:0.1, D:0.2)root;'))
        clusters = Tree2Tax().named_clusters(tree, 0.05)
        self.assertSameClusters([['A'],['B'],['D']], clusters)
        assert_equals(_('Root.1 Root.2 Root.3'), [c.name() for c in clusters])
         
    def testClusterNamingWithBootstraps(self):
        tree = TreeNode.read(StringIO("((F:20, ((A:11, B:12):10, (H:8, D:9):3):20)'0.7:G':30)root;"))
        clusters = Tree2Tax().named_clusters(tree, 40)
        self.assertSameClusters([['F'], _('A B D H')], clusters)
        assert_equals(_('G.2 G.1'), [c.name() for c in clusters])
         
    def testClusterNamingConventionsWithSomeUnnamed(self):
        tree = TreeNode.read(StringIO('((((A:11, B:12):10, D:9):20, F:20)G:30)root;'))
        clusters = Tree2Tax().named_clusters(tree, 0.05) #i.e. everything is a separate cluster
        self.assertSameClusters([['A'],['B'],['D'],['F']], clusters)
        assert_equals(['G.1', 'G.2', 'G.3', 'G.4'], [c.name() for c in clusters])
 
         
class TestTree2TaxNamedClusterSets:
    def assertSameClusterSets(self, expected, observed):
        assert_equals(len(expected), len(observed))
        for i, e in enumerate(expected):
            assert_equals(e[0], observed[i].threshold) #same thresholds
            self.assertSameClusters(e[1], observed[i].clusters)
     
    def assertSameClusters(self, expected, observed):
        name_clusters = [sorted([n.name for n in named_cluster.tips]) for named_cluster in observed]
        expected_name_clusters = [sorted(node_names) for node_names in expected]
        #print "named clusters: %s" % name_clusters
        assert_equals(expected_name_clusters, name_clusters)
         
    def testSimple(self):
        tree = TreeNode.read(StringIO('((A:0.11, B:0.12)C:0.1, D:0.2)root;'))
        clusters = Tree2Tax().named_clusters_for_several_thresholds(tree, [0.25])
        self.assertSameClusterSets([[0.25,[['A','B'],['D']]]], clusters)
 
    def testSimpleTwice(self):
        tree = TreeNode.read(StringIO('((A:0.11, B:0.12)C:0.1, D:0.2)root;'))
        clusters = Tree2Tax().named_clusters_for_several_thresholds(tree, [0.25, 0.25])
        self.assertSameClusterSets([[0.25,[['A','B'],['D']]], [0.25,[['A','B'],['D']]]], clusters)
         
    def testNormalType(self):
        tree = TreeNode.read(StringIO('((A:0.11, B:0.12)C:0.1, D:0.2)root;'))
        clusters = Tree2Tax().named_clusters_for_several_thresholds(tree, [0.25, 0.05])
        self.assertSameClusterSets([[0.05,[['A'],['B'],['D']]], [0.25,[['A','B'],['D']]]], clusters)
        assert_equals(_('C.1 C.2 Root'), [c.name() for c in clusters[0].clusters])
         
    def testOppositeSorting(self):
        tree = TreeNode.read(StringIO('((A:0.11, B:0.12)C:0.1, D:0.2)root;'))
        clusters = Tree2Tax().named_clusters_for_several_thresholds(tree, [0.05, 0.25])
        self.assertSameClusterSets([[0.05,[['A'],['B'],['D']]], [0.25,[['A','B'],['D']]]], clusters)
        assert_equals(_('C.1 C.2 Root'), [c.name() for c in clusters[0].clusters])
         
    def testNaming(self):
        tree = TreeNode.read(StringIO('((F:20, ((A:11, B:12):10, (H:8, D:9):3):20)G:30)root;'))
        clusters = Tree2Tax().named_clusters_for_several_thresholds(tree, [40, 25])
        self.assertSameClusterSets([[25,[['F'], _('A B'), _('D H')]], [40,[['F'], _('A B D H')]]], clusters)
        assert_equals(_('G.3 G.1 G.2'), [c.name() for c in clusters[0].clusters])
        assert_equals(_('G.2 G.1'), [c.name() for c in clusters[1].clusters])
 
    def testTipToCluster(self):
        tree = TreeNode.read(StringIO('((F:20, ((A:11, B:12):10, (H:8, D:9):3):20)G:30)root;'))
        clusters = Tree2Tax().named_clusters_for_several_thresholds(tree, [40, 25])
        self.assertSameClusterSets([[25,[['F'], _('A B'), _('D H')]], [40,[['F'], _('A B D H')]]], clusters)
        assert_equals(_('G.3 G.1 G.2'), [c.name() for c in clusters[0].clusters])
        assert_equals(_('G.2 G.1'), [c.name() for c in clusters[1].clusters])
         
        tip = tree.find('F')
        assert_equals('G.3', clusters[0].tip_to_cluster(tip).name())
        assert_equals('G.2', clusters[1].tip_to_cluster(tip).name())
 
        tip = tree.find('D')
        assert_equals('G.2', clusters[0].tip_to_cluster(tip).name())
        assert_equals('G.1', clusters[1].tip_to_cluster(tip).name())
 
class TestNamedCluster:
    def testCondensedName(self):
        nc = NamedCluster('c__Halo; o__fu', ['notips'], None)
        assert_equals('cHalo.ofu', nc.condensed_name())
         
        nc.cluster_number = 6
        assert_equals('cHalo.ofu.6', nc.condensed_name())
         
        nc.taxonomy = 'c__Halo'
        assert_equals('cHalo.6', nc.condensed_name())
     


from skbio.tree import TreeNode
from sets import Set
import logging
import re
import IPython

class TaxonomyFunctions:
    @staticmethod
    def condense(taxonomy_string):
        '''e.g. if taxonomy is c__Halobacteria; o__Halobacteriales; f__MSP41
        then return 'cHalobacteria.oHalobacteriales.fMSP41' We cannot omit the 
        c, o or f because
        in some rare cases two taxonomic levels have the same name except for their
        level e.g. c__Gemmatimonadetes; o__Gemmatimonadetes in GreenGenes 
        2013_08'''
        splits = taxonomy_string.split('; ')
        
        # get rid of f__ prefixes etc.
        regex = re.compile(r'^.__')
        splits2 = []
        for s in splits:
            reg = regex.match(s)
            if reg:
                splits2.append(s[0]+s[reg.end():])
            else:
                logging.debug("Found unexpected form for taxonomy in %s", str(s))
                splits2.append(s)
        return '.'.join(splits2)
    
    @staticmethod
    def missing_taxonomy(tree, descendent_node, ancestral_node):
        '''given a tree, and two nodes where one is an ancestor of another,
        return a condensed taxonomy list representing the 
        taxonomic information that is contained in nodes between the descendent 
        and ancestral nodes, except for the first one encountered
        '''
        to_return = []
        
        current = descendent_node
        while current != ancestral_node:
            current = current.parent
            if current is None:
                raise Exception("descendent and ancestral nodes do not appear to be related as expected in #missing_taxonomy")
            
            if current != ancestral_node:
                tax = TaxonomyFunctions().taxonomy_from_node_name(current.name)
                if tax:
                    to_return.append(TaxonomyFunctions().condense(tax))

        # return in descending order
        return [r for r in reversed(to_return)]
            
    @staticmethod
    def taxonomy_from_node_name(node_name):
        '''return the taxonomy incorporated at a particular node, or None
        if it does not encode any taxonomy'''
        
        def isFloat(s):
            try:
                float(s)
                return True
            except ValueError:
                return False
        
        if node_name is None:
            return None
        elif isFloat(node_name):
            # no name, just a bootstrap
            return None
        else:
            bootstrap_regex = re.compile(r'[\d\.]+:(.*)')
            reg = bootstrap_regex.match(node_name)
            if reg:
                # bootstrap in name
                return reg.groups(0)[0]
            else:
                # bootstrap not in name
                return node_name
    

class NamedCluster:
    def __init__(self, taxonomy, tips, lca_node):
        self.taxonomy = taxonomy
        self.tips = tips
        self.lca_node = lca_node
        self.cluster_number = None
        
    def name(self):
        if self.cluster_number:
            return "%s.%s" % (self.taxonomy, self.cluster_number)
        else:
            return self.taxonomy
        
    def __str__(self):
        return "%s %s: %s" % (self.taxonomy, self.cluster_number, self.tips)
    
    def condensed_name(self):
        '''e.g. if taxonomy is c__Halobacteria; o__Halobacteriales; f__MSP41
        then return 'cHalobacteria.oHalobacteriales.fMSP41' (cluster number
        is added too, if that is present. We cannot omit the c, o or f because
        in some rare cases two taxonomic levels have the same name except for their
        level e.g. c__Gemmatimonadetes; o__Gemmatimonadetes in GreenGenes 
        2013_08'''
        joined = TaxonomyFunctions.condense(self.taxonomy)
        if self.cluster_number:
            return "%s.%s" % (joined, self.cluster_number)
        else:
            return joined
        
        
class ThresholdAndClusters:
    def __init__(self, threshold, clusters):
        self.threshold = threshold
        self.clusters = clusters
        
    def tip_name_to_cluster(self, tip_name):
        key = tip_name
        try:
            # cached already?
            return self._tip_to_cluster[key]
        except AttributeError:
            # no dice. Have to create the hash
            self._tip_to_cluster = {}
            for cluster in self.clusters:
                for datip in cluster.tips:
                    if self._tip_to_cluster.has_key(datip.name):
                        logging.warn("Unexpectedly found multiple leaf nodes with the same name, undefined behaviour possibly imminent: %s" % tip_name)
                    else:
                        self._tip_to_cluster[datip.name] = cluster
        return self._tip_to_cluster[key]
        
    def tip_to_cluster(self, tip):
        '''return the NamedCluster to which the specified tip belongs'''
        return self.tip_name_to_cluster(tip.name)
        
                    
    def each_tip(self):
        '''iterate over all the tips in all clusters'''
        for cl in self.clusters:
            for tip in cl.tips:
                yield tip
            

class Tree2Tax:
    def named_clusters_for_several_thresholds(self, original_tree, thresholds):
        '''Given a list of thresholds, return a iterable of ThresholdAndClusters
        where the clustering has been done iteratively, providing a consistent
        taxonomic annotation scheme'''
        original_tree.assign_ids()
        tree = original_tree.copy() #this copy gets destructively pruned as part of the algorithm
        
        # sort from smallest to largest because smaller distance thresholds
        # need to be applied before larger thresholds
        sorted_thresholds = sorted(thresholds)
        if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug("Found thresholds %s", str(sorted_thresholds))
        
        to_return = []

        # skbio seems to be quite slow running repeated find_by_id. Do some caching 
        # to speed things up
        id_to_node = {}
        for node in original_tree.postorder(include_self=True):
            id_to_node[node.id] = node
            
        clades_to_distances = {}
        for threshold in sorted_thresholds:
            if logging.getLogger().isEnabledFor(logging.INFO): logging.info("Clustering on threshold %s.." % str(threshold))
            
            clades_to_distances = self.destructively_cluster_tree(tree, threshold, clades_to_distances)
            
            clusters = []
            if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug(tree.ascii_art())
            cluster_lcas = tree.tips()
            
            # So that several clades don't get named the same thing
            # it get a bit complex when several nodes are annotated as having the sam
            # taxonomy (it happens..., in gg at least). So group by taxonomy
            # rather than node ID
            taxonomy_to_named_nodes = {} 
            
            # when everything is in 1 cluster (doesn't happen in practice I suspect)
            # but there is a unit test..
            if len(tree) == 0:
                cl = NamedCluster('Root', original_tree.tips(), original_tree)
                cl.cluster_number = ''
                clusters = [cl]
    
            else:
                for lca in cluster_lcas:
                    orig = id_to_node[lca.id]
                    
                    # Find the closest ancestral named node (where ancestral does not include self)
                    current = orig
                    
                    #in a greengenes file, tips have names, but don't count these 
                    #because they are not taxonomy but rather prokMSA IDs
                    if current.is_tip(): current = current.parent
                    
                    # Keep proceeding up the tree until a node with taxonomy is found.
                    # If the tree has bootstraps, then current.name is a float (in string form)
                    # Ignore these floats because they aren't named taxonomy
                    while current.parent and (current.name is None or 
                         TaxonomyFunctions.taxonomy_from_node_name(current.name) is None):
                        current = current.parent
                        
                    if current.parent is None:
                        taxonomy = 'Root'
                    else:
                        taxonomy = TaxonomyFunctions.taxonomy_from_node_name(current.name)
                        
                    if orig.is_tip():
                        named_cluster = NamedCluster(taxonomy, [orig], orig)
                    else:
                        named_cluster = NamedCluster(taxonomy, list(orig.tips()), orig)
                        
                    clusters.append(named_cluster)
                    
                    try:
                        taxonomy_to_named_nodes[taxonomy].append(named_cluster)
                    except KeyError:
                        taxonomy_to_named_nodes[taxonomy] = [named_cluster]
                
                if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug(taxonomy_to_named_nodes)
                # Assign cluster numbers by decreasing numbers of tips i.e. for each
                # taxonomy, the cluster with the most sequences is given the number 1,
                # second most abundant is number 2, etc.
                for taxonomy, named_clusters in taxonomy_to_named_nodes.items():
                    if len(named_clusters) > 1:
                        number = 1
                        # sort twice (stupid python). Consider number of tips
                        # to be more important than the minimum node ID. Need
                        # to sort repeatably for the purposes of testing
                        sorts1 = sorted(named_clusters, key = lambda c: min([tip.id for tip in c.tips]))
                        for clade in sorted(sorts1, reverse = True, key = lambda c: len(c.tips)):
                            clade.cluster_number = number
                            number += 1
                            
            threshold_and_clusters = ThresholdAndClusters(threshold, clusters) 
            to_return.append(threshold_and_clusters)
            
        return to_return
        
        
    def named_clusters(self, original_tree, threshold):
        '''cluster a tree given a threshold tree_distance'''
        array = self.named_clusters_for_several_thresholds(original_tree, [threshold])
        return array[0].clusters
    
    def destructively_cluster_tree(self, tree, threshold, clades_to_distances = {}):
        '''Given a tree, collapse tips using complete linkage such that 
        all sequences within the collapsed clade have at most the threshold 
        tree distance (and cluster as much as possible). Return hash of 
        tip.id => farthest distance found from that root that still obeyed
        the threshold'''
        
        tips_to_evaluate = Set(tree.tips())
        if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug("Parsed tree with %s tips: %s" % (tree.count(tips=True), [t.name for t in tips_to_evaluate]))
        if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug(tree.ascii_art())
        
        while len(tips_to_evaluate) > 0:
            tip = tips_to_evaluate.pop() #raises KeyError when set is empty, jumping out of the while loop
            if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug("Evaluating tip %s/%s" % (tip.id, tip.name))
            
            # If this tip has only 1 sibling, then evaluate whether it and its sibling can be collapsed according to the threshold
            if tip.parent is not None: #if we aren't at the root
                bros = [n for n in tip.parent.children if n != tip and n.is_tip()]
                
                if len(bros) == 1:
                    # The sibling node is a tip
                    bro = bros[0]
                    distance_to_bro = tip.length + bro.length
                    if tip.id in clades_to_distances: distance_to_bro += clades_to_distances[tip.id]
                    if bro.id in clades_to_distances: distance_to_bro += clades_to_distances[bro.id]
                    if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug("found bro %s/%s at distance %s" % (bro.id, bro.name, distance_to_bro))
                    if distance_to_bro <= threshold:
                        # cluster these two tips
                        # remove tip and bro from the tree
                        # add the parent node to the set so it can possibly can merge with further tips or parents that become tips
                        if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug("Below threshold, removing these nodes")
                        if bro in tips_to_evaluate: tips_to_evaluate.remove(bro)
                        parental = tip.parent
                        tips_to_evaluate.add(parental)
                        if parental.remove(bro) is not True: raise Exception("Programming error")
                        if parental.remove(tip) is not True: raise Exception("Programming error")
                        
                        # Set the distance of this node to be the maximal distance to below
                        # a la complete linkage clustering
                        max_dist = 0.0
                        for n in [tip, bro]:
                            try:
                                # if n was originally an internal node,
                                # there should be a clades_to_distances entry 
                                # associated. Distance is then that distance
                                # plus the distance between n and parental
                                d = clades_to_distances[n.id] + n.length
                            except KeyError:
                                d = n.length
                            if d > max_dist: max_dist = d
                        clades_to_distances[parental.id] = max_dist
                        if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug("Setting maximal distance at %s" % max_dist)

                        if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug(tree.ascii_art())
                    else:
                        # Don't merge, but remove bro node from the list to visit,
                        # if it isn't already removed
                        if bro in tips_to_evaluate: tips_to_evaluate.remove(bro)
        
                # else there is multiple siblings. Don't need to add this back into the set again because it will be rediscovered later.
                else:
                    if logging.getLogger().isEnabledFor(logging.DEBUG): logging.debug("The sibling tree is not a single tip, ignoring this node")
                    
        return clades_to_distances


#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    'scikit-bio>=0.2.2,<1.0',
    'tempdir',
]

test_requirements = [
    'nose'
]

exec(open('tree2tax/version.py').read()) # loads __version__

setup(
    name='tree2tax',
    version=__version__,
    description="Automatic taxonomy through consistent application of tree-based thresholding",
    long_description=readme + '\n\n' + history,
    author="Ben Woodcroft",
    author_email='b.woodcroft near uq.edu.au',
    url='https://github.com/wwood/tree2tax',
    packages=[
        'tree2tax',
    ],
    package_dir={'tree2tax':
                 'tree2tax'},
    include_package_data=True,
    install_requires=requirements,
    license="LGPL3",
    zip_safe=False,
    keywords='tree2tax',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    scripts=[
        'bin/autotaxonomy',
        'bin/percent_identity_vs_tree_distance',
        'bin/threshold_estimator',
        'bin/tree2tax'
    ]
)



__author__ = "Ben Woodcroft"
__copyright__ = "Copyright 2014-2015"
__credits__ = ["Ben Woodcroft"]
__license__ = "LGPL3"
__maintainer__ = "Ben Woodcroft"
__email__ = "b.woodcroft near uq.net.au"
__status__ = "Development"

from .version import __version__



import re
import logging
from sets import Set

from skbio.tree import TreeNode



class CladeDistanceSet:
    def __init__(self):
        self.clade_distances = []
    
    def append(self, clade_distance):
        self.clade_distances.append(clade_distance)
        
    def average_distance(self):
        sum([cd.distance for cd in self.clade_distances]) / len(self.clade_distancess)

class CladeDistance:
    def __init__(self, parent_node, daughter_node1, daughter_node2, distance):
        self.parent_node = parent_node
        self.daughter_node1 = daughter_node1
        self.daughter_node2 = daughter_node2
        self.distance = distance
        
class Stack:
    def __init__(self):
        self.__storage = []

    def isEmpty(self):
        return len(self.__storage) == 0

    def push(self,p):
        self.__storage.append(p)

    def pop(self):
        return self.__storage.pop()

class ThresholdFinder:
    def find_examples(self, tree, upper_prefix, lower_prefix):
        '''return a CladeDistanceSet of pairs of clades from the lower_prefix
        rank that are sisters in the upper_prefix rank'''
        to_return = []
        
        # get a list of nodes that are at the upper threshold, by
        # descending the tree
        upper_prefix_regex = re.compile(r'(; ){0,1}%s__' % upper_prefix)
        upper_nodes = []
        for node in tree.non_tips():
            if node.name and upper_prefix_regex.match(node.name):
                upper_nodes.append(node)
        logging.debug("Found %s upper nodes" % len(upper_nodes))
        
        # For each of these upper class prefixes    
        # find all the lower level nodes
        lower_prefix_regex = re.compile(r'(; ){0,1}%s__' % lower_prefix)    
        for unode in upper_nodes:
            lower_nodes = []
            for node in unode.non_tips(True):
                if node.name and lower_prefix_regex.match(node.name):
                    lower_nodes.append(node)
            logging.debug("Lower nodes: %s" % lower_nodes)
                
            # if there is only 1, go to next upper because there are no pairs
            if len(lower_nodes) <= 1: continue
        
            # for each of the lower level nodes, find the maximal distance to the tips
            # special cases - if one of the lower nodes is the LCA,
            # then the max distance of that LCA cannot be derived from a 
            # tip in the other
            lower_max_distances = self.unique_best_distance(lower_nodes)
                                   
            # make/add to a list of all the pairwise distances,
            for i, first in enumerate(lower_nodes):
                for j, second in enumerate(lower_nodes):
                    if i <= j: continue #upper right triangle of distance matrices only
                    
                    # by taking each pair of lower nodes, 
                    # distance is distance of node1 to lca + distance of node2 to lca
                    # + max distance underneath of each node
                    
                    # there is currently (v 0.2.2) a bug in scikit-bio's implementation
                    # of distance - so workaround here
                    # https://github.com/biocore/scikit-bio/issues/807
                    lca = self.my_lca(tree, first, second)
                    distance = first.accumulate_to_ancestor(lca) + \
                        second.accumulate_to_ancestor(lca) + \
                        lower_max_distances[i] + \
                        lower_max_distances[j]
                        
                    to_return.append(CladeDistance(unode, first, second, distance))
        
        # return the list of distances
        return to_return
    
    def unique_best_distance(self, nodes):
        '''given a set of nodes, return an array of max distances to tips
        that are descendants of each node. However, the 'max tip' cannot be a
        descendant of any other node. Assumes no tips are counted as a clade'''
        to_return = []
        
        annotated_nodes = Set(nodes)
            
        for node in nodes:
            # Make a set for fast membership querying

            stack = Stack()
            max_distance = 0.0
            
            # initialise stack with first descendents
            for d in node.children:
                stack.push(d) 
                
            while True:
                try:
                    d = stack.pop()
                except IndexError:
                    break #stack empty, we are done
                
                if d not in annotated_nodes: # don't descend when we run into a different clade
                    if d.is_tip():
                        distance = d.accumulate_to_ancestor(node)
                        if distance > max_distance: 
                            max_distance = distance
                    else:
                        [stack.push(desc) for desc in d.children]

                
            to_return.append(max_distance)
        return to_return
    
    def my_lca(self, tree, first, second):
        '''v0.2.2 of skbio is buggy, here's a quick fix'''
        first_ancestors = Set([first] + first.ancestors())
        curr = second
        while curr:
            if curr in first_ancestors:
                return curr
            else:
                curr = curr.parent
                
        raise Exception("No common ancestors found, whack - are they from the same tree?") 
    
    
